#include <getopt.h>
#include <string.h>
#include <stdio.h>
#include "machine.h"
#include "libpperf.h"

#ifndef PERF_DOT_CONF
#define PERF_DOT_CONF "/etc/pperf.conf"
#endif

/* Table of events. */
static Events *events;
/* for 'machine()' */
static char arch[128];
static float mhz;
static int num_events;

void print_usage (void);
void print_event (void);
void pperf_setup (void);
  
void getargs( int argc, char **argv, char **runfile, 
	      int *event0, int *event1, int *cpl0, 
	      int *cpl1, int *verb, float *t_step )
{
  int c, i;
  int file_flag = 1;
  int digit_optind = 0;
  int setup_flag = 0;
  char display[128];

  /* Get architecture information. */

  machine(arch,&mhz,&num_events,0);

  /* Set the defaults. */

  if (!strcmp(arch,"i586"))
    {
      events = p5_events;
      *event0 = events[P5_FLOPS].id;
      *event1 = events[P5_MEM_DATA_READ].id;
    }

  if (!strcmp(arch,"i686"))
    {
      events = p6_events;
      *event0 = events[P6_FLOPS].id;
      *event1 = events[P6_DATA_MEM_REFS].id;
    }

  *cpl0 = 2;
  *cpl1 = 2;
  *verb = LONG;
  *t_step = 0.0;

  /* Parse the command line arguments. */

  while (1)
    {
      int this_option_optind = optind ? optind : 1;
      int option_index = 0;
      static struct option long_options[] = 
      {
	{"help", 0, 0, 'h'},
	{"list", 0, 0, 'l'},
	{"cpl0", 1, 0, 0},
	{"cpl1", 1, 0, 0},
	{"event0", 1, 0, '0'},
	{"event1", 1, 0, '1'},
	{"display", 1, 0, 'd'},
	{"setup", 0, 0, 'S'},
	{"step", 1, 0, 's'},
	{0, 0, 0, 0}
      };

      c = getopt_long (argc, argv, "lhS0:1:d:s:",
		       long_options, &option_index);
      if (c == -1)
	break;

      switch (c)
	{
	case 0:
	  if (!strcmp(long_options[option_index].name,"cpl0"))
	    {
	      *cpl0 = atoi(optarg);
	    }
	  if (!strcmp(long_options[option_index].name,"cpl1"))
	    {
	      *cpl1 = atoi(optarg);
	    }

	case ' ':
	  /* If this isn't in here, ' ' is treated as '0' which
	   *     messes up case '0'.
	   */
	  break;

        case '0': /* Event Selector 0 */
	  *event0 = atoi(optarg);
          break;
          
        case '1': /* Event Selector 1 */
	  *event1 = atoi(optarg);
          break;
          
	case 'h':
	  print_usage();
	  exit(0);
	  
	case 'l':
	  print_event();
	  exit(0);

	case 'd':
	  strcpy(display,optarg);
	  break;

	case 's':
	  sscanf(optarg,"%f",t_step);
	  break;

	case 'S':
	  pperf_setup();
	  exit(0);

	case '?':
	  break;
	  
	default:
	  printf ("?? getopt returned character code 0%o ??\n", c);
	}
    }

  /* The program (and it's arguments) which we will run. */

  if (optind < argc)
    {
      i = 0;
      while (optind < argc)
	{
	  runfile[i] = argv[optind++];
	  i++;
	}
      /* Last Argument - so execvp doesn't bark. */
      runfile[i] = NULL;  
    }
  else
    {
      fprintf(stderr,"pperf: No program specified.\n\n");
      print_usage();
      exit(0);
    }

  /* Sanity check- have been asked for non-existent events? */

  if (*event0 < 0 || *event0 > num_events)
    {
      fprintf(stderr,"error: Fatal error.  No event %d defined for %s.\n",
	      *event0,arch);
      print_usage();
      exit(0);
    }

  if (*event1 < 0 || *event1 > num_events)
    {
      fprintf(stderr,"error: Fatal error.  No event %d defined for %s.\n",
	      *event1,arch);
      print_usage();
      exit(0);
    }

  /* Have we been given a valid CPL? */

  if (*cpl0 < 0 || *cpl0 > 3)
    {
      fprintf(stderr,"error: Fatal error. Invalid CPL0=%d. CPL0=[0,1,2,3].\n",
	      *cpl0);
      print_usage();
      exit(0);
    }

  if (*cpl1 < 0 || *cpl1 > 3)
    {
      fprintf(stderr,"error: Fatal error. Invalid CPL1=%d. CPL1=[0,1,2,3].\n",
	      *cpl1);
      print_usage();
      exit(0);
    }

  /* Make sure time step is reasonable.  t_step was set to 0.0
  *     as a default so it can be used to see if it is requested
  *     inside pperf.c */

  if (*t_step < 0.02 && *t_step != 0.0)
    {
      fprintf(stderr,"warning: Time step out of bounds.\n");
      fprintf(stderr,"\tTime step must be greater than 0.02 sec (20 ms).\n");
      fprintf(stderr,"\tResetting time step to 1 sec.\n");
      *t_step = 1.0;
    }

  /* What kind of detail in the output is requested? */

  if ( !strcmp(display,"s") )
    {
      *verb = SHORT;
    }

  if ( !strcmp(display,"l") )
    {
      *verb = LONG;
    }

  if ( !strcmp(display,"d") )
    {
      *verb = DETAILED;
    }
}

void pperf_setup ( void )
{
  char buf[128];
  FILE *fp;
  int cpu_speed_found = 0;
  long pos;

  if ( (fp = fopen(PERF_DOT_CONF,"r+")) == NULL )
    {
      if ( (fp = fopen(PERF_DOT_CONF,"w+")) == NULL )
	{
	  fprintf(stderr,
		  "error: Cannot open %s for writing.\n",PERF_DOT_CONF);
	  fprintf(stderr,
		  "\tMust be root to run setup.\n");
	}
      else 
	{
	  fprintf(fp,"CPU_SPEED=%.2f\n",mhz);
	}
    }
  else
    {
      rewind(fp);
      pos = ftell(fp);
      while ( fgets(buf,sizeof(buf),fp) != NULL )
	{
	  if ( strstr(buf,"CPU_SPEED") != NULL && strstr(buf,"#") == NULL )
	    {
	      fseek(fp,pos,SEEK_SET);
	      fprintf(fp,"CPU_SPEED=%.2f\n",mhz);
	      cpu_speed_found = 1;
	    }
	  pos = ftell(fp);
	}
      if (cpu_speed_found == 0)
	{
	  fprintf(fp,"CPU_SPEED=%.2f\n",mhz);
	}
    }
  fclose(fp);
}  
  
void print_usage(void)
{
  printf("       
       pperf  [-0  |  --event0]  event  [-1  |  --event1] event
       [--cpl0] [0,1,2,3] [--cpl1]  [0,1,2,3]  [-d  |  --display]
       [l,s,d]  [-s  |  --step]  [time (sec)] [-l | --list] [-h |
       --help] [-S | --setup] program [program arguments]

       To see a list of valid events for use with --es0 and --es1
       use pperf -l.

       pperf  -  use hardware specific registers in the Pentium
       or Pentium Pro processor to  monitor  various  aspects  of
       program performance.

       For additional information see the pperf(2) manual page.

");
}

void print_event ( void )
{
  int i;
  
  for (i=0; i<num_events; i++)
    {
      printf("%4d  %s\n",events[i].id,events[i].desc);
    }
}
