#include <stdio.h>
#include <string.h>
#include <sys/fcntl.h>

#define _USE_BSD
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include "libpperf.h"

#define RETSIGTYPE void
/* A Pointer to a signal handler.  */
typedef RETSIGTYPE (*sighandler) ();

#define PPERF_NOT_STARTED 0
#define PPERF_STARTED 1

#define ARCH_NOT_DEFINED 0
#define ARCH_DEFINED 1

#define COMPUTE_SPEED 0
#define READ_SPEED 1

/* Global scope to libpperf.c only. */
static int num_events;
static Events *events;
static int e0, e1;
long long Clock;
int start_flag = PPERF_NOT_STARTED;
int arch_flag = ARCH_NOT_DEFINED;
int g_status = NO_ERROR;

pid_t child_pid = CHILD_NOT_STARTED;
pid_t parent_pid;

int start_counters( int event0, int cpl0, int event1, int cpl1 )
{
  int i, flag0, flag1;
  unsigned int event_sel0[2], event_sel1[2];
  unsigned int clear[2];
  int status = NO_ERROR;

  unsigned int umask0 = 0x0, umask1 = 0x0;

  int fd;

  /* Have we determined the architecture yet? */

  if (arch_flag == ARCH_NOT_DEFINED)
    {
      machine(pperf_arch,&pperf_mhz,&num_events,READ_SPEED);
      arch_flag = ARCH_DEFINED;
    }

  /* Sanity check- have been asked for non-existent events? */

  if (event0 < 0 || event0 > num_events)
    {
      fprintf(stderr,"error: Fatal error.  No event %d defined for %s.\n",
	      event0,pperf_arch);
      exit(0);
    }

  if (event1 < 0 || event1 > num_events)
    {
      fprintf(stderr,"error: Fatal error.  No event %d defined for %s.\n",
	      event1,pperf_arch);
      exit(0);
    }

  /* Have we been given a valid CPL? */

  if (cpl0 < 0 || cpl0 > 3)
    {
      fprintf(stderr,"error: Fatal error. Invalid CPL0=%d. CPL0=[0,1,2,3].\n",
	      cpl0);
      exit(0);
    }

  if (cpl1 < 0 || cpl1 > 3)
    {
      fprintf(stderr,"error: Fatal error. Invalid CPL1=%d. CPL1=[0,1,2,3].\n",
	      cpl1);
      exit(0);
    }

  /* Open /dev/msr */
      
  if ( (fd = open("/dev/msr", O_WRONLY)) == -1 ) 
    {
      fprintf(stderr,"warning: Open /dev/msr WRONLY failed.\n");
      fprintf(stderr,"\tProbable cause: msr.o module not loaded.\n");
      status = DEV_MSR_NOT_OPENED;
      g_status = status;
    }

  /* Write to /dev/msr on Pentium */

  /* Pack event_sel with the right stuff.
   * 
   * Pentium:
   *     
   *     Both events are selected with a single event selection
   *     register (CESR).
   *     
   *     3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 
   *     1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
   *     
   *     x x x x x x P [ C ] [    E    ] x x x x x x P [ C ] [    E    ]
   *     x x x x x x C [ C ] [    S    ] x x x x x x C [ C ] [    S    ]
   *     x x x x x x 1 [ 1 ] [    1    ] x x x x x x 0 [ 0 ] [    0    ]
   *     
   *     xxx: RESERVED
   *     PC1, PC0: 0=(Event Increment) 1=(Event Overflow)
   *     CC1, CC0: CPL Level to Monitor, possibilities are
   *               000 = Count Nothing
   *               001 = Count Event while CPL = 0,1,2
   *               010 = Count Event while CPL = 3
   *               011 = Count Event while CPL = 0,1,2,3
   *               100 = Count Nothing
   *               101 = Count Clocks while CPL = 0,1,2
   *               110 = Count Clocks while CPL = 3
   *               111 = Count Clocks while CPL = 0,1,2,3
   *               Here we only use 000, 001, 010 and 011.
   *     ES1, ES0: Event selection code, see events.c
   *     
   */

  clear[0] = 1 << 31;
  clear[1] = 0;  

  if (!strcmp(pperf_arch,"i586"))
    {
      events = p5_events;
      
      event_sel0[1] = (events[event0].opcode) ^ (cpl0 << 6) ^ (0 << 9) ^
	(events[event1].opcode << 16) ^ (cpl1 << 22) ^ (0 << 25);

      /* Zero both 40-bit counters. */

      lseek(fd, 0x12, 0);
      write(fd, clear, 16);
      lseek(fd, 0x13, 0);
      write(fd, clear, 16);

      /* Write out CESR. */

      lseek(fd, 0x11, 0);
      write(fd, event_sel0, 8);
    }

  /* Write to /dev/msr on PPro/PII */

  /* Pentium Pro/II:
   *     
   *     Each event is specified seperately in an EVENTSEL register.
   *     
   *     3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1 
   *     1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
   *     
   *     [      C      ] I x x I B O O U [      U      ] [      E      ]
   *     [      M      ] N x x N P D S S [      M      ] [      V      ]
   *     [      K      ] V x x T 0 M P P [      K      ] [      T      ]
   *     
   *     xxx: RESERVED
   *     CMK: Counter Mask
   *     INV: Inverts CMK, 0=(Inverted) 1=(Non-Inverted)
   *     INT: APIC Signalling of counter overflow 1=(Enable) 0=(Disable)
   *     BP0: Signal Overflow via BP0 pin 1=(Enable) 0=(Disable)
   *     ODM: Occurance/Duration Mode Select 1=(Occurance) 0=(Duration)
   *     OSP: Count events while CPL=0 1=(Enable) 0=(Disable)
   *     USP: Count events while CPL=1,2,3 1=(Enable) 0=(Disable)
   *     UMK: Unit Mask (Used for MESI protocol events)
   *     EVT: Event selection code, see events.c
   */

  if (!strcmp(pperf_arch,"i686"))
    {
      events = p6_events;

      /* If we are checking an L2 event that requires a mask, then set
       *     the mask to count all MESI events.
       */

      if (event0 == 10 || event0 == 11 || event0 == 12 || event0 == 17)
	umask0 = 0xf;

      if (event1 == 10 || event1 == 11 || event1 == 12 || event1 == 17)
	umask1 = 0xf;

      /* To use CPL = 0,1,2,3 consistently, we need to reverse bits
       *     for CPL = 1 and CPL = 2
       */      

      if (cpl0 == 1 || cpl0 == 2)
	cpl0 = ~cpl0 & 0x3;

      if (cpl1 == 1 || cpl1 == 2)
	cpl1 = ~cpl1 & 0x3;

      event_sel0[1] = (events[event0].opcode) ^ (umask0 << 8) ^ (cpl0 << 16) ^
	(0 << 18) ^ (0 << 19) ^ (0 << 20) ^ (1 << 22) ^ (0 << 24);

      event_sel1[1] = (events[event1].opcode) ^ (umask1 << 8) ^ (cpl1 << 16) ^
	(0 << 18) ^ (0 << 19) ^ (0 << 20) ^ (1 << 22) ^ (0 << 24);

      /* Clear COUNTER0 */

      lseek(fd, 0xc1, 0);
      write(fd, clear, 8);

      /* Clear COUNTER1 */

      lseek(fd, 0xc2, 0);
      write(fd, clear, 16);

      /* Write out EVENTSEL0. */

      lseek(fd, 0x186, 0);
      write(fd, event_sel0, 8);

      /* Write out EVENTSEL1. */

      lseek(fd, 0x187, 0);
      write(fd, event_sel1, 8);
    }

  /* To look at what values we passed. */
#ifdef DEBUG
  printf("Event0 (Passed %d):\n",event0);
  printf("\t%d %d 0x%x %s %s\n",events[event0].id,events[event0].rflag,
	 events[event0].opcode,events[event0].name,events[event0].desc);

  printf("Event1 (Passed %d):\n",event1);
  printf("\t%d %d 0x%x %s %s\n",events[event1].id,events[event1].rflag,
	 events[event1].opcode,events[event1].name,events[event1].desc);

  printf("CPL0: %d  CPL1: %d \n",cpl0,cpl1);

  printf("es0: 0x%x 0x%x\n",event_sel0[0],event_sel0[1]);
  printf("es1: 0x%x 0x%x\n",event_sel1[0],event_sel1[1]);
#endif

  close(fd);

  /* Save these so other libpperf.c functions can use them. */

  e0 = event0;
  e1 = event1;

  /* Save starting TSC counter. */

  Clock = rdtsc();

  /* Let's let other libpperf functions know we have started profiling. */

  start_flag = PPERF_STARTED;

  return(status);
}

long long rdpmc ( unsigned int reg )
{
  unsigned int tmp[2];

  __asm__ ("rdpmc"
	   : "=a" (tmp[1]), "=d" (tmp[0])
	   : "c" (reg) );

  return ( ((long long)(tmp[0] & 0xff)) << 32 | tmp[1] );
}

long long rdtsc ( void )
{
  unsigned int tmp[2];

  __asm__ ("rdtsc"
	   : "=a" (tmp[1]), "=d" (tmp[0])
	   : "c" (0x10) );
  
  return ( ((long long)tmp[0] << 32 | tmp[1]) );
}

void Check_Prof_Start ( char *lib_func )
{
  if (start_flag == PPERF_NOT_STARTED)
    {
      fprintf(stderr,"error: Fatal error.  Profiling not initialized.\n");
      fprintf(stderr,"\tIssue the start_counters command prior to %s.\n",
	      lib_func);
      exit(0);
    }
}

int read_counters ( long long *count0, long long *count1, double *timer, 
		    long long *clocks )
{
  unsigned int tmp[2];
  int fd;
  int status = NO_ERROR;

  /* If msr.o is not installed, just leave. */

  if (g_status)
    return(g_status);

  /* Make sure we have actually started profiling. */

  Check_Prof_Start(__FUNCTION__);

  /* Read in the clock ticks. */

  *clocks = rdtsc();
  *timer = (double) *clocks/pperf_mhz/1.0e6;

  if (!strcmp(pperf_arch,"i586"))
    {
      /* This has not been tested yet!!! */
      /* If this does work- no need to check arch. */
      /* FIXME:  Does this work with Pentium???? */
      /*
      *count0 = rdpmc(0x0);
      *count1 = rdpmc(0x1);
      */
      /* Can't get rdpmc to work, defaulting to this: */
      if ( (fd=open("/dev/msr", O_RDONLY)) == -1 )
	{
	  fprintf(stderr,
		  "warning: Cannot open /dev/msr O_RDONLY.\n");
	  fprintf(stderr,"\tProbable cause: msr.o module not loaded.\n");
	  status = DEV_MSR_NOT_OPENED;
	}

      lseek(fd, 0x12, 0);
      read(fd, tmp, 8);
      *count0 =( ((long long)(tmp[0] & 0xff)) << 32 | tmp[1] );
      lseek(fd, 0x13, 0);
      read(fd, tmp, 8);
      *count1 =( ((long long)(tmp[0] & 0xff)) << 32 | tmp[1] );
      close(fd);
    }

  if (!strcmp(pperf_arch,"i686"))
    {
      *count0 = rdpmc(0x0);
      *count1 = rdpmc(0x1);
    }

  return(status);
}

int dread_counters ( double *count0, double *count1, double *timer, 
		    double *clocks )
{
  unsigned int tmp[2];
  int fd;
  int status = NO_ERROR;

  /* If msr.o is not installed, just leave. */

  if (g_status)
    return(g_status);

  /* Make sure we have actually started profiling. */

  Check_Prof_Start(__FUNCTION__);

  /* Read in the clock ticks. */

  *clocks = rdtsc();
  *timer = (double) *clocks/pperf_mhz/1.0e6;


  if (!strcmp(pperf_arch,"i586"))
    {
      /* This has not been tested yet!!! */
      /* If this does work- no need to check arch. */
      /* FIXME:  Does this work with Pentium???? */
      /*
      *count0 = rdpmc(0x0);
      *count1 = rdpmc(0x1);
      */
      /* Can't get rdpmc to work, defaulting to this: */
      if ( (fd=open("/dev/msr", O_RDONLY)) == -1 )
	{
	  fprintf(stderr,
		  "warning: Cannot open /dev/msr O_RDONLY.\n");
	  fprintf(stderr,"\tProbable cause: msr.o module not loaded.\n");
	  status = DEV_MSR_NOT_OPENED;
	}

      lseek(fd, 0x12, 0);
      read(fd, tmp, 8);
      *count0 = (double) ( ((long long)(tmp[0] & 0xff)) << 32 | tmp[1] );
      
      lseek(fd, 0x13, 0);
      read(fd, tmp, 8);
      *count1 = (double) ( ((long long)(tmp[0] & 0xff)) << 32 | tmp[1] );
      close(fd);
    }

  if (!strcmp(pperf_arch,"i686"))
    {
      *count0 = (double) rdpmc(0x0);
      *count1 = (double) rdpmc(0x1);
    }

  return(status);
}

int print_counters ( int verb )
{
  long long count0, count1, clocks;
  double timer;
  double c0, c1;
  long long C_diff;
  float T_diff;
  int status = NO_ERROR;

  /* If msr.o is not installed, just leave. */

  if (g_status)
    return(g_status);

  /* Make sure we have actually started profiling. */

  Check_Prof_Start(__FUNCTION__);

  status = read_counters(&count0, &count1, &timer, &clocks);

  C_diff = clocks-Clock;
  T_diff = (float) C_diff/pperf_mhz/1.0e6;
  c0 = (double) count0;
  c1 = (double) count1;

  switch(verb)
    {
    case SHORT:
      printf("%s  %lld  %s  %lld  %f  %lld\n",
	     events[e0].name,count0,events[e1].name,count1,T_diff,C_diff);
      break;
    case LONG:
      printf("
--------------------- libpperf ---------------------
Profiling Summary Since Last start_counters Call:
 
   Event Selected                         Counts
     %-25s  %16.0lf
     %-25s  %16.0lf

   Clock Ticks:  %lld
   Time:  %.3f (sec)
^^^^^^^^^^^^^^^^^^^^^ libpperf ^^^^^^^^^^^^^^^^^^^^^
",events[e0].name,c0,events[e1].name,c1,C_diff,T_diff);
      break;
    case DETAILED:
      printf("
--------------------- libpperf ---------------------
Profiling Summary Since Last start_counters Call:

   Event Selected                         Counts
     %-25s  %16.0lf
     %-25s  %16.0lf

   Clock Ticks:  %lld
   Time:  %.3f (sec)

Events Details:

   Event 0:
      Name: %s
      Description: %s
      Event Number: %d
      Event Opcode: 0x%x
      Event rcode: 0x%x

   Event 0:
      Name: %s
      Description: %s
      Event Number: %d
      Event Opcode: 0x%x
      Event rcode: 0x%x

Timing Details:

   Cumulative Clocks Since Reset: %lld
   Cumulative Time Since Reset: %.4lf (sec)

Architecture Details:

   Architecture: %s
   CPU Speed: %.2f (MHz)
^^^^^^^^^^^^^^^^^^^^^ libpperf ^^^^^^^^^^^^^^^^^^^^^
",events[e0].name,c0,events[e1].name,c1,C_diff,T_diff,events[e0].name,
	     events[e0].desc,events[e0].id,events[e0].opcode,events[e0].rflag,
	     events[e1].name,events[e1].desc,events[e1].id,events[e1].opcode,
	     events[e1].rflag,clocks,timer,pperf_arch,pperf_mhz);
      break;
    default:
      fprintf(stderr,
	      "warning: Invalid argument to print_counters, using SHORT.\n");
      printf("%s  %lld  %s  %lld  %f  %lld\n",
	     events[e0].name,count0,events[e1].name,count1,timer,clocks);
      break;
    }

  fflush(stdout);
  return(status);
}

double rtime ( void )
{
  long long tsc;

  /* Read the TSC. */

  tsc = rdtsc();

  /* If this is our first call this program, define the architecture. */

  if (arch_flag == ARCH_NOT_DEFINED)
    {
      machine(pperf_arch,&pperf_mhz,&num_events,READ_SPEED);
      arch_flag = ARCH_DEFINED;
    }

  /* Return the system time since reboot.  (Actually, the system time
   *     (read TSC) _could_ be overwritten.  But I have chosen not to
   *     do this, so unless the user has overwritten the TSC this will
   *     work.  Question:  with a 2^64 TSC _WHY_ would one overwrite
   *     the TSC???
   */

  return( (double) tsc/pperf_mhz/1.0e6 );
}

void pstatus ( int err )
{
  switch(err)
    {
    case NO_ERROR:
      fprintf(stderr,"*** No errors reported.\n\n");
      break;
    case DEV_MSR_NOT_OPENED:
      fprintf(stderr,"*** /dev/msr could not be accessed.\n");
      fprintf(stderr,"*** Probable cause: msr.o module not installed.\n");
      fprintf(stderr,"*** Returned counter results are meaningless.\n\n");
      break;
    case CHILD_PROC_STARTED:
      fprintf(stderr,"*** Already started a child process.\n\n");
      break;
    case CHILD_NOT_STARTED:
      fprintf(stderr,"*** No child process to terminate.\n\n");
      break;
    case TIMING_TOO_SMALL:
      fprintf(stderr,"*** Request for time slices smaller than\n");
      fprintf(stderr,"*** context switching time.  (About 20 ms).\n\n");
      break;
    default:
      fprintf(stderr,"*** Error code not recognized.\n\n");
      break;
    }
}

void child_handler ( int j )
{
  printf("**************************\n");
  kill(child_pid,SIGTERM);
}  

int start_print_counters ( float sleep_time, int verb )
{
  long long count0, count1, clocks;
  double timer;
  int status = NO_ERROR;
  unsigned long usec;

  /* If msr.o is not installed, just leave. */

  if (g_status)
    return(g_status);

  /* If we've already started a child, then quit. */

  if (child_pid != CHILD_NOT_STARTED)
    {
      fprintf(stderr,"warning: Already started a child process.\n");
      fprintf(stderr,"\tNo further children will be created.\n");
      return(CHILD_PROC_STARTED);
    }

  parent_pid = getpid();

  /* Allows profiling in increments of time smaller than 1 sec.
   *     However, if the time requested is finer grained than
   *     the context switch time (about .02 sec (20 ms) then
   *     it is moot, so default to 20000 microseconds.
   */

  usec = (unsigned long)(sleep_time*1000000);

  if (usec < 20000)
    {
      usec = 20000;
      status = TIMING_TOO_SMALL;
    }

  if ( (child_pid = fork()) < 0 )
    {
      fprintf(stderr,"error: Fatal error.  Cannot fork.\n");
      exit(0);
    }

  if (child_pid == 0)
    {
      while (getppid() == parent_pid)
	{
	  print_counters(verb);
	  usleep(usec);
	  /*
	  sleep(sleep_time);
	  */
	}
      stop_print_counters();
    }

  return (status);
}

int stop_print_counters ( void )
{
  if (child_pid == CHILD_NOT_STARTED)
    {
      fprintf(stderr,"warning: No child process to terminate.\n");
      return (CHILD_NOT_STARTED);
    }
  else
    {
      kill(child_pid,SIGTERM);
      child_pid = CHILD_NOT_STARTED;
      return (NO_ERROR);
    }
}

/* FORTRAN wrappers.  Single underscore. */

int Fortran(start_counters)( int *event0, int *cpl0, int *event1, int *cpl1 )
{
  return((int)start_counters(*event0,*cpl0,*event1,*cpl1));
}

long long Fortran(rdpmc)( unsigned int *reg )
{
  return((long long)rdpmc(*reg));
}

long long Fortran(rdtsc)( void )
{
  return((long long)rdtsc());
}

int Fortran(read_counters)( long long *count0, long long *count1, 
			    double *timer, long long *clocks )
{
  return((int)read_counters(count0,count1,timer,clocks));
}

int Fortran(dread_counters)( double *count0, double *count1, double *timer, 
			     double *clocks )
{
  return((int)dread_counters(count0,count1,timer,clocks));
}

int Fortran(print_counters)( int *verb )
{
  return((int)print_counters(*verb));
}

double Fortran(rtime)( void )
{
  return((double)rtime());
}

void Fortran(pstatus)( int *err )
{
  pstatus(*err);
}

int Fortran(start_print_counters)( float *sleep_time, int *verb )
{
  return((int)start_print_counters(*sleep_time,*verb));
}

int Fortran(stop_print_counters)( void )
{
  return((int)stop_print_counters());
}

/* FORTRAN wrappers.  Two underscores. */

int Fortran2(start_counters)( int *event0, int *cpl0, int *event1, int *cpl1 )
{
  return((int)start_counters(*event0,*cpl0,*event1,*cpl1));
}

long long Fortran2(rdpmc)( unsigned int *reg )
{
  return((long long)rdpmc(*reg));
}

long long Fortran2(rdtsc)( void )
{
  return((long long)rdtsc());
}

int Fortran2(read_counters)( long long *count0, long long *count1, 
			    double *timer, long long *clocks )
{
  return((int)read_counters(count0,count1,timer,clocks));
}

int Fortran2(dread_counters)( double *count0, double *count1, double *timer, 
			     double *clocks )
{
  return((int)dread_counters(count0,count1,timer,clocks));
}

int Fortran2(print_counters)( int *verb )
{
  return((int)print_counters(*verb));
}

double Fortran2(rtime)( void )
{
  return((double)rtime());
}

void Fortran2(pstatus)( int *err )
{
  pstatus(*err);
}

int Fortran2(start_print_counters)( float *sleep_time, int *verb )
{
  return((int)start_print_counters(*sleep_time,*verb));
}

int Fortran2(stop_print_counters)( void )
{
  return((int)stop_print_counters());
}
