#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

/* Define available events in Event structure. */

typedef struct {
  int id;
  int rflag;
  unsigned int opcode;
  char name[40];
  char desc[120];
} Events;

static Events p6_events[] = {
  /* Data Cache Unit (DCU) */
  {0,2,0x43,"data_mem_refs","all memory references, cachable and non"},
  {1,2,0x45,"dcu_lines_in","total lines allocated in the DCU"},
  {2,2,0x46,"dcu_m_lines_in","number of M state lines allocated in DCU"},
  {3,2,0x47,"dcu_m_lines_out","number of M lines evicted from the DCU"},
  {4,2,0x48,"dcu_miss_outstanding","number of cycles while DCU miss outstanding"},
  /* Intruction Fetch Unit (IFU) */
  {5,2,0x80,"ifu_ifetch","number of non/cachable instruction fetches"},
  {6,2,0x81,"ifu_ifetch_miss","number of instruction fetch misses"},
  {7,2,0x85,"itlb_miss","number of ITLB misses"},
  {8,2,0x86,"ifu_mem_stall","cycles instruction fetch pipe is stalled"},
  {9,2,0x87,"ild_stall","cycles instruction length decoder is stalled"},
  /* L2 Cache */
  {10,2,0x28,"l2_ifetch","number of L2 instruction fetches"},
  {11,2,0x29,"l2_ld","number of L2 data loads"},
  {12,2,0x2a,"l2_st","number of L2 data stores"},
  {13,2,0x24,"l2_lines_in","number of allocated lines in L2"},
  {14,2,0x26,"l2_lines_out","number of recovered lines from L2"},
  {15,2,0x25,"l2_m_lines_inm","number of modified lines allocated in L2"},
  {16,2,0x27,"l2_m_lines_outm","number of modified lines removed from L2"},
  {17,2,0x2e,"l2_rqsts","number of L2 requests"},
  {18,2,0x21,"l2_ads","number of L2 address strobes"},
  {19,2,0x22,"l2_dbus_busy","number of cycles data bus was busy"},
  {20,2,0x23,"l2_dmus_busy_rd","cycles data bus was busy in xfer from L2 to CPU"},
  /* External Bus Logic (EBL) */
  {21,2,0x62,"bus_drdy_clocks","number of clocks DRDY is asserted"},
  {22,2,0x63,"bus_lock_clocks","number of clocks LOCK is asserted"},
  {23,2,0x60,"bus_req_outstanding","number of outstanding bus requests"},
  {24,2,0x65,"bus_tran_brd","number of burst read transactions"},
  {25,2,0x66,"bus_tran_rfo","number of read for ownership transactions"},
  {26,2,0x67,"bus_trans_wb","number of write back transactions"},
  {27,2,0x68,"bus_tran_ifetch","number of instruction fetch transactions"},
  {28,2,0x69,"bus_tran_inval","number of invalidate transactions"},
  {29,2,0x6a,"bus_tran_pwr","number of partial write transactions"},
  {30,2,0x6b,"bus_trans_p","number of partial transactions"},
  {31,2,0x6c,"bus_trans_io","number of I/O transactions"},
  {32,2,0x6d,"bus_trans_def","number of deferred transactions"},
  {33,2,0x6e,"bus_tran_burst","number of burst transactions"},
  {34,2,0x70,"bus_tran_any","number of all transactions"},
  {35,2,0x6f,"bus_tran_mem","number of memory transactions"},
  {36,2,0x64,"bus_data_rcv","bus cycles this processor is receiving data"},
  {37,2,0x61,"bus_bnr_drv","bus cycles this processor is driving BNR pin"},
  {38,2,0x7a,"bus_hit_drv","bus cycles this processor is driving HIT pin"},
  {39,2,0x7b,"bus_hitm_drv","bus cycles this processor is driving HITM pin"},
  {40,2,0x7e,"bus_snoop_stall","cycles during bus snoop stall"},
  /* Floating Point Unit (FPU) */
  {41,0,0xc1,"comp_flop_ret","number of computational FP operations retired"},
  {42,0,0x10,"flops","number of computational FP operations executed"},
  {43,1,0x11,"fp_assist","number of FP execptions handled by microcode"},
  {44,1,0x12,"mul","number of multiplies"},
  {45,1,0x13,"div","number of divides"},
  {46,0,0x14,"cycles_div_busy","cycles divider is busy"},
  /* Memory Ordering */
  {47,2,0x03,"ld_blocks","number of store buffer blocks"},
  {48,2,0x04,"sb_drains","number of store buffer drain cycles"},
  {49,2,0x05,"misalign_mem_ref","number of misaligned data memory references"},
  /* Instruction Decoding and Retirement */
  {50,2,0xc0,"inst_retired","number of instructions retired"},
  {51,2,0xc2,"uops_retired","number of UOPs retired"},
  {52,2,0xd0,"inst_decoder","number of instructions decoded"},
  /* Interrupts */
  {53,2,0xc8,"hw_int_rx","number of hardware interrupts received"},
  {54,2,0xc6,"cycles_int_masked","cycles interrupts are disabled"},
  {55,2,0xc7,"cycles_int_pending_and_masked","cycles interrupts are disabled with pending interrupts"},
  /* Branches */
  {56,2,0xc4,"br_inst_retired","number of branch instructions retired"},
  {57,2,0xc5,"br_miss_pred_retired","number of mispredicted bracnhes retired"},
  {58,2,0xc9,"br_taken_retired","number of taken branches retired"},
  {59,2,0xca,"br_miss_pred_taken_ret","number of taken mispredictions branches retired"},
  {60,2,0xe0,"br_inst_decoded","number of branch instructions decoded"},
  {61,2,0xe2,"btb_misses","number of branches that miss the BTB"},
  {62,2,0xe4,"br_bogus","number of bogus branches"},
  {63,2,0xe6,"baclears","number of times BACLEAR is asserted"},
  /* Stalls */
  {64,2,0xa2,"resource_stalls","cycles during resource related stalls"},
  {65,2,0xd2,"partial_rat_stalls","cycles or events for partial stalls"},
  /* Segment Register Loads */
  {66,2,0x06,"segment_reg_loads","number of segment register loads"},
  /* Clocks */
  {67,2,0x79,"cpu_clk_unhalted","clocks processor is not halted"}};

static Events p5_events[] = {
  {0,2,0x00,"mem_data_read","data read hits"},
  {1,2,0x01,"mem_data_write","data write hits"},
  {2,2,0x02,"tlb_miss","data TLB misses"},
  {3,2,0x03,"mem_data_rm","data read misses"},
  {4,2,0x04,"mem_data_wm","data write misses"},
  {5,2,0x05,"write_hit_me","writes (hits) to M/E"},
  {6,2,0x06,"data_cache_wb","data cache lines written back"},
  {7,2,0x07,"ext_snoops","external snoops"},
  {8,2,0x08,"data_cache_snoop_hits","data cache snoop hits"},
  {9,2,0x09,"mem_accs_both_pipes","memory accesses in both pipes"},
  {10,2,0x0A,"bank_conflicts","bank conflicts"},
  {11,2,0x0B,"misal_mem_ref","misaligned data memory references"},
  {12,2,0x0C,"code_read","code reads"},
  {13,2,0x0D,"code_tlb_miss","code TLB misses"},
  {14,2,0x0E,"code_cache_miss","code cache misses"},
  {15,2,0x0F,"seg_reg_load","segment register loaded"},
  {16,2,0x10,"seg_desc_cache_accs","segment descriptor cache accesses"},
  {17,2,0x11,"seg_desc_cache_hit","segment descriptor cache hits"},
  {18,2,0x12,"branches","branches"},
  {19,2,0x13,"btb_hits","BTB hits"},
  {20,2,0x14,"branch_or_btb_hit","taken branches or BTB hits"},
  {21,2,0x15,"pipeline_flush","pipeline flushes"},
  {22,2,0x16,"ins_exe_b_pipes","instructions executed in both pipes"},
  {23,2,0x17,"ins_exe_v_pipe","instructions executed in V-pipe"},
  {24,2,0x18,"clks_bus_cycle","clocks while bus cycle in progress"},
  {25,2,0x19,"pipe_stl_fwb","pipe stalled by full write buffers"},
  {26,2,0x1A,"pipe_stl_wdr","pipe stalled by waiting for data reads"},
  {27,2,0x1B,"pipe_stl_wme","pipe stalled by writes to M/E"},
  {28,2,0x1C,"locked_bus","locked bus cycles"},
  {29,2,0x1D,"io_read_write","I/O read or write cycles"},
  {30,2,0x1E,"non-cache_mem_ref","non-cacheable memory references"},
  {31,2,0x1F,"pipe_stl_agi","pipeline stalled by AGI"},
  {32,2,0x22,"flops","floating-point operations"},
  {33,2,0x23,"brk_dr0","breakpoint matches on DR0"},
  {34,2,0x24,"brk_dr1","breakpoint matches on DR1"},
  {35,2,0x25,"brk_dr2","breakpoint matches on DR2"},
  {36,2,0x26,"brk_dr3","breakpoint matches on DR3"},
  {37,2,0x27,"hdw_int","hardware interrupts"},
  {38,2,0x28,"mem_read_write_hit","data reads or writes"},
  {39,2,0x29,"mem_read_write_miss","data read/write misses"}};

/* Event defines for easier use of events. */

/* Pentium Events */

#define P5_MEM_DATA_READ 0 
#define P5_MEM_DATA_WRITE 1 
#define P5_TLB_MISS 2 
#define P5_MEM_DATA_RM 3 
#define P5_MEM_DATA_WM 4 
#define P5_WRITE_HIT_ME 5 
#define P5_DATA_CACHE_WB 6 
#define P5_EXT_SNOOPS 7 
#define P5_DATA_CACHE_SNOOP_HITS 8 
#define P5_MEM_ACCS_BOTH_PIPES 9 
#define P5_BANK_CONFLICTS 10 
#define P5_MISAL_MEM_REF 11 
#define P5_CODE_READ 12 
#define P5_CODE_TLB_MISS 13 
#define P5_CODE_CACHE_MISS 14 
#define P5_SEG_REG_LOAD 15 
#define P5_SEG_DESC_CACHE_ACCS 16 
#define P5_SEG_DESC_CACHE_HIT 17 
#define P5_BRANCHES 18 
#define P5_BTB_HITS 19 
#define P5_BRANCH_OR_BTB_HIT 20 
#define P5_PIPELINE_FLUSH 21 
#define P5_INS_EXE_B_PIPES 22 
#define P5_INS_EXE_V_PIPE 23 
#define P5_CLKS_BUS_CYCLE 24 
#define P5_PIPE_STL_FWB 25 
#define P5_PIPE_STL_WDR 26 
#define P5_PIPE_STL_WME 27 
#define P5_LOCKED_BUS 28 
#define P5_IO_READ_WRITE 29 
#define P5_NON_CACHE_MEM_REF 30 
#define P5_PIPE_STL_AGI 31 
#define P5_FLOPS 32 
#define P5_BRK_DR0 33 
#define P5_BRK_DR1 34 
#define P5_BRK_DR2 35 
#define P5_BRK_DR3 36 
#define P5_HDW_INT 37 
#define P5_MEM_READ_WRITE_HIT 38 
#define P5_MEM_READ_WRITE_MISS 39 

/* Pentium Pro/PII Events */

#define P6_DATA_MEM_REFS 0 
#define P6_DCU_LINES_IN 1 
#define P6_DCU_M_LINES_IN 2 
#define P6_DCU_M_LINES_OUT 3 
#define P6_DCU_MISS_OUTSTANDING 4 
#define P6_IFU_IFETCH 5 
#define P6_IFU_IFETCH_MISS 6 
#define P6_ITLB_MISS 7 
#define P6_IFU_MEM_STALL 8 
#define P6_ILD_STALL 9 
#define P6_L2_IFETCH 10 
#define P6_L2_LD 11 
#define P6_L2_ST 12 
#define P6_L2_LINES_IN 13 
#define P6_L2_LINES_OUT 14 
#define P6_L2_M_LINES_INM 15 
#define P6_L2_M_LINES_OUTM 16 
#define P6_L2_RQSTS 17 
#define P6_L2_ADS 18 
#define P6_L2_DBUS_BUSY 19 
#define P6_L2_DMUS_BUSY_RD 20 
#define P6_BUS_DRDY_CLOCKS 21 
#define P6_BUS_LOCK_CLOCKS 22 
#define P6_BUS_REQ_OUTSTANDING 23 
#define P6_BUS_TRAN_BRD 24 
#define P6_BUS_TRAN_RFO 25 
#define P6_BUS_TRANS_WB 26 
#define P6_BUS_TRAN_IFETCH 27 
#define P6_BUS_TRAN_INVAL 28 
#define P6_BUS_TRAN_PWR 29 
#define P6_BUS_TRANS_P 30 
#define P6_BUS_TRANS_IO 31 
#define P6_BUS_TRANS_DEF 32 
#define P6_BUS_TRAN_BURST 33 
#define P6_BUS_TRAN_ANY 34 
#define P6_BUS_TRAN_MEM 35 
#define P6_BUS_DATA_RCV 36 
#define P6_BUS_BNR_DRV 37 
#define P6_BUS_HIT_DRV 38 
#define P6_BUS_HITM_DRV 39 
#define P6_BUS_SNOOP_STALL 40 
#define P6_COMP_FLOP_RET 41 
#define P6_FLOPS 42 
#define P6_FP_ASSIST 43 
#define P6_MUL 44 
#define P6_DIV 45 
#define P6_CYCLES_DIV_BUSY 46 
#define P6_LD_BLOCKS 47 
#define P6_SB_DRAINS 48 
#define P6_MISALIGN_MEM_REF 49 
#define P6_INST_RETIRED 50 
#define P6_UOPS_RETIRED 51 
#define P6_INST_DECODER 52 
#define P6_HW_INT_RX 53 
#define P6_CYCLES_INT_MASKED 54 
#define P6_CYCLES_INT_PENDING_AND_MASKED 55 
#define P6_BR_INST_RETIRED 56 
#define P6_BR_MISS_PRED_RETIRED 57 
#define P6_BR_TAKEN_RETIRED 58 
#define P6_BR_MISS_PRED_TAKEN_RET 59 
#define P6_BR_INST_DECODED 60 
#define P6_BTB_MISSES 61 
#define P6_BR_BOGUS 62 
#define P6_BACLEARS 63 
#define P6_RESOURCE_STALLS 64 
#define P6_PARTIAL_RAT_STALLS 65 
#define P6_SEGMENT_REG_LOADS 66 
#define P6_CPU_CLK_UNHALTED 67 

/* C Code Declarations */

extern int start_counters( int event0, int cpl0, int event1, 
			   int cpl1 );
extern int read_counters ( long long *count0, long long *count1, 
			   double *timer, long long *clocks );
extern int dread_counters ( double *count0, double *count1, 
		  	    double *timer, double *clocks );
extern long long rdpmc ( unsigned int reg );
extern long long rdtsc ( void );
extern int print_counters ( int verb );
extern double rtime ( void );
extern void pstatus ( int err );
extern int start_print_counters ( float sleep_time, int verb );
extern int stop_print_counters ( void );

/* External Fortran linkage */

#define Fortran(x) x##_
#define Fortran2(x) x##__

/* Some FORTRAN compilers look for one underscore, some two. */

/* FORTRAN Code Declarations */

extern int Fortran(start_counters)( int *event0, int *cpl0, int *event1, 
				    int *cpl1 );
extern int Fortran(read_counters)( long long *count0, long long *count1, 
				   double *timer, long long *clocks );
extern int Fortran(dread_counters)( double *count0, double *count1, 
				    double *timer, double *clocks );
extern long long Fortran(rdpmc)( unsigned int *reg );
extern long long Fortran(rdtsc)( void );
extern int Fortran(print_counters) ( int *verb );
extern double Fortran(rtime)( void );
extern void Fortran(pstatus)( int *err );
extern int Fortran(start_print_counters)( float *sleep_time, int *verb );
extern int Fortran(stop_print_counters)( void );

extern int Fortran2(start_counters)( int *event0, int *cpl0, int *event1, 
				    int *cpl1 );
extern int Fortran2(read_counters)( long long *count0, long long *count1, 
				   double *timer, long long *clocks );
extern int Fortran2(dread_counters)( double *count0, double *count1, 
				    double *timer, double *clocks );
extern long long Fortran2(rdpmc)( unsigned int *reg );
extern long long Fortran2(rdtsc)( void );
extern int Fortran2(print_counters) ( int *verb );
extern double Fortran2(rtime)( void );
extern void Fortran2(pstatus)( int *err );
extern int Fortran2(start_print_counters)( float *sleep_time, int *verb );
extern int Fortran2(stop_print_counters)( void );

/* Global scope variables so we don't have to run function
 *     machine more than once, which would slow down the
 *     execution time.
 */
char pperf_arch[128];
float pperf_mhz;

/* This is for print_counters.  Amount of information printed. */
#define SHORT    0
#define LONG     1
#define DETAILED 2

/* This is for start_counters and get_cpu_time */
#define MHZ_COMP 0
#define MHZ_READ 1

/* Return codes. */
#define NO_ERROR 0
#define DEV_MSR_NOT_OPENED 1
#define CHILD_NOT_STARTED -1
#define CHILD_PROC_STARTED 2
#define TIMING_TOO_SMALL 3








